@@ -15,21 +15,22 @@
 from __future__ import print_function
 
 import getopt
-import re
 import os
+import re
 import sys
 import uuid
 
-from ovs.db import error
 import ovs.db.idl
 import ovs.db.schema
-from ovs.db import data
 import ovs.db.types
 import ovs.ovsuuid
 import ovs.poller
 import ovs.stream
 import ovs.util
+from ovs.db import data
+from ovs.db import error
 from ovs.fatal_signal import signal_alarm
+
 import six
 
 
@@ -288,10 +289,7 @@ def idltest_find_simple2(idl, i):
 
 
 def idltest_find_simple3(idl, i):
-    for row in six.itervalues(idl.tables["simple3"].rows):
-        if row.name == i:
-            return row
-    return None
+    return next(idl.index_equal("simple3", "simple3_by_name", i), None)
 
 
 def idl_set(idl, commands, step):
@@ -333,7 +331,12 @@ def idl_set(idl, commands, step):
             if args[1] == "b":
                 s.b = args[2] == "1"
             elif args[1] == "s":
-                s.s = args[2]
+                if six.PY2:
+                    s.s = args[2].decode('utf-8')
+                else:
+                    s.s = args[2].encode(sys.getfilesystemencoding(),
+                                         'surrogateescape') \
+                                 .decode('utf-8', 'replace')
             elif args[1] == "u":
                 s.u = uuid.UUID(args[2])
             elif args[1] == "r":
@@ -573,6 +576,8 @@ def do_idl(schema_file, remote, *commands):
     else:
         schema_helper.register_all()
     idl = ovs.db.idl.Idl(remote, schema_helper)
+    if "simple3" in idl.tables:
+        idl.index_create("simple3", "simple3_by_name")
 
     if commands:
         error, stream = ovs.stream.Stream.open_block(
@@ -843,7 +848,7 @@ def main(argv):
             sys.stderr.write("%s: \"%s\" requires at least %d arguments but "
                              "only %d provided\n"
                              % (ovs.util.PROGRAM_NAME, command_name,
-                                n_args, len(args)))
+                                n_args[0], len(args)))
             sys.exit(1)
     elif type(n_args) == int:
         if len(args) != n_args:
